"""
Hoppr stage output Rich layout
"""
from __future__ import annotations

from rich.console import Console
from rich.containers import Lines
from rich.layout import Layout
from rich.panel import Panel
from rich.progress import BarColumn, Progress, TextColumn
from rich.text import Text

from hoppr import __version__

BORDER_LINES = 2
HEADER_SIZE = 3
REMOVE_LINES = BORDER_LINES + HEADER_SIZE

console = Console()


class HopprJobsPanel(Panel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress(
        "{task.description}",
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    )

    def __init__(self) -> None:
        super().__init__(
            renderable=self.progress_bar,
            title="[bold blue]Jobs",
            title_align="left",
            border_style="purple",
        )

    def add_task(self, description: str, **fields) -> int:
        """
        Add a task to be tracked in the `jobs` side panel

        Args:
            description (str): A description of the task

        Returns:
            int: ID of the added task
        """
        return self.progress_bar.add_task(description, start=False, **fields)


class HopprOutputPanel(Panel):
    """
    Customized Rich Text box to simulate a console
    """

    lines = Lines()

    def __init__(self) -> None:
        super().__init__(
            renderable=self.lines,
            border_style="purple",
            expand=False,
            title_align="left",
        )

    def print(self, line: str, style: str = "") -> None:
        """
        Write a message to the console output panel

        Args:
            line (str): Message to write
            style (str): Style to apply to message string
        """
        self.lines.append(Text(line, style))

        # Subtract static lines of Panel boxes to get displayed height of text
        if len(self.lines) > console.height - REMOVE_LINES:
            self.lines.pop(index=0)


class HopprProgressPanel(Panel):  # pylint: disable=too-few-public-methods
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress()

    def __init__(self) -> None:
        super().__init__(
            renderable=self.progress_bar,
            title="[bold blue]Progress",
            title_align="left",
            border_style="purple",
        )

        self.progress_bar.add_task(description="All Jobs")


class HopprLayout(Layout):
    """
    Layout of the Hoppr console application
    """

    name: str = "root"
    jobs_panel = HopprJobsPanel()
    output_panel = HopprOutputPanel()
    overall_progress = HopprProgressPanel()

    def __init__(self, title: str = f"Hoppr v{__version__}") -> None:
        super().__init__()

        self.split(Layout(name="header", size=HEADER_SIZE), Layout(name="main"))
        self["main"].split_row(Layout(name="side"), Layout(name="console", ratio=2))
        self["side"].split(Layout(name="jobs"), Layout(name="progress", size=3))

        # Initialize header
        header = Text(text=title, style="bold blue", justify="center")
        self["header"].update(renderable=Panel(renderable=header, border_style="purple"))

        # Initialize jobs side bar panel
        self["jobs"].update(renderable=self.jobs_panel)

        # Initialize overall progress side bar
        self["progress"].update(renderable=self.overall_progress)

        # Initialize main body panel
        self["console"].update(renderable=self.output_panel)

    def add_job(self, description: str, **fields) -> int:
        """
        Add a job to the `jobs` side panel

        Args:
            description (str): Description of the job to add
        """
        return self.jobs_panel.add_task(description, **fields)

    def print(self, line: str, style: str = "") -> None:
        """
        Write a message to the console output panel

        Args:
            line (str): Message to write
        """
        self.output_panel.print(line, style)



"""
`bundle` subcommand for `hoptctl`
"""
from __future__ import annotations

from pathlib import Path
from typing import Optional

from typer import Argument, Option, Typer

from hoppr import __version__, main

app = Typer(
    context_settings={"help_option_names": ['-h', '--help']},
    help="Run the stages specified in the transfer config file on the content specified in the manifest",
    invoke_without_command=True,
    no_args_is_help=True,
    rich_markup_mode="markdown",
    subcommand_metavar="",
)


@app.callback()
def bundle(
    # pylint: disable=unused-argument
    # pylint: disable=too-many-arguments
    manifest_file: Path = Argument(
        "manifest.yml",
        help="Path to manifest file",
        expose_value=True,
    ),
    credentials_file: Optional[Path] = Option(
        None,
        "-c",
        "--credentials",
        help="Specify credentials config for services",
        envvar="HOPPR_CREDS_CONFIG",
    ),
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
    log_file: Optional[Path] = Option(
        None,
        "-l",
        "--log",
        help="File to which log will be written",
        envvar="HOPPR_LOG_FILE",
    ),
    verbose: bool = Option(
        False,
        "-v",
        "--debug",
        "--verbose",
        help="Enable debug output",
    ),
    strict_repos: bool = Option(
        True,
        "--strict/--no-strict",
        help="Utilize only manifest repositories for package collection",
        envvar="HOPPR_STRICT_REPOS",
    ),
    create_attestations: bool = Option(
        False,
        "-a",
        "--attest",
        help="Generate in-toto attestations",
        envvar="HOPPR_ATTESTATION",
    ),
    functionary_key_path: Optional[Path] = Option(
        None,
        "-fk",
        "--functionary-key",
        help="Path to key used to sign in-toto layout",
        envvar="HOPPR_FUNCTIONARY_KEY",
    ),
    functionary_key_prompt: bool = Option(
        False,
        "-p",
        "--prompt",
        help="Prompt user for project owner key's password",
        envvar="HOPPR_PROJECT_OWNER_KEY_PROMPT",
    ),
    functionary_key_password: str = Option(
        None,
        "-fk-pass",
        "--project-owner-key-password",
        help="Password for project owner key",
        envvar="HOPPR_PROJECT_OWNER_KEY_PASSWORD",
    ),
    previous_delivery: Optional[Path] = Option(
        None,
        "-pd",
        "--previous-delivery",
        help="Path to manifest or tar bundle representing a previous delivery",
        envvar="HOPPR_PREVIOUS_DELIVERY",
    ),
):  # pragma: no cover
    """
    Run the stages specified in the transfer config
    file on the content specified in the manifest
    """
    main.bundle(**locals())



"""
Top-level CLI application
"""
from __future__ import annotations

import ctypes
import os
import sys

from pathlib import Path
from platform import python_version

from typer import Option, Typer, secho, style

from hoppr import __version__, main
from hoppr.cli import bundle, generate, merge

# Windows flags and types
NT_ENABLE_ECHO_INPUT = 0b0100
NT_ENABLE_LINE_INPUT = 0b0010
NT_ENABLE_PROCESSED_INPUT = 0b0001
NT_CONSOLE_FLAGS = NT_ENABLE_ECHO_INPUT | NT_ENABLE_LINE_INPUT | NT_ENABLE_PROCESSED_INPUT
NT_STD_OUTPUT_HANDLE = ctypes.c_uint(-11)

# Enable ANSI processing on Windows systems
if sys.platform == "win32":  # pragma: no cover
    nt_kernel = ctypes.WinDLL(name="kernel32.dll")

    nt_kernel.SetConsoleMode(nt_kernel.GetStdHandle(NT_STD_OUTPUT_HANDLE), NT_CONSOLE_FLAGS)


app = Typer(
    name="hopctl",
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Collect, process, & bundle your software supply chain",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
)

app.add_typer(typer_instance=bundle.app)
app.add_typer(typer_instance=generate.app)
app.add_typer(typer_instance=merge.app)


# Aliases for deprecated commands to preserve backward compatibility
generate_layout = app.command(
    name="generate-layout",
    deprecated=True,
    rich_help_panel="Deprecated",
    help="See `hopctl generate layout` subcommand",
)(generate.layout)


generate_schemas = app.command(
    name="generate-schemas",
    deprecated=True,
    rich_help_panel="Deprecated",
    help="See `hopctl generate schemas` subcommand",
)(generate.schemas)


@app.command()
def validate(
    input_files: list[Path],
    credentials_file: Path = Option(
        None,
        "-c",
        "--credentials",
        help="Specify credentials config for services",
        envvar="HOPPR_CREDS_CONFIG",
    ),
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
):  # pragma: no cover
    """
    Validate multiple manifest files for schema errors
    """
    main.validate(input_files, credentials_file, transfer_file)


@app.command()
def version():
    """
    Print version information for `hoppr`
    """
    # TTY terminals and MacOS Terminal.app do not support ANSI multibyte characters. Print low-resolution art instead
    suffix = ".ansi" if sys.stdout.isatty() and os.getenv("TERM_PROGRAM") != "Apple_Terminal" else ".ascii"
    hippo = (Path(main.__file__).parent / "resources" / "hoppr-hippo").with_suffix(suffix)

    secho(hippo.read_bytes())

    secho(f"{style(text='Hoppr Framework Version', fg='green')} : {__version__}")
    secho(f"{style(text='Python Version         ', fg='green')} : {python_version()}")


__all__ = ["app"]



"""
`merge` subcommand for `hoptctl`
"""
from __future__ import annotations

import time

from pathlib import Path

import typer

from pydantic import ValidationError
from rich.console import Console, Group
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn
from rich.table import Table
from typer import BadParameter, Option, Typer

from hoppr.cli.layout import HopprJobsPanel, HopprLayout
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import CycloneDXBaseModel, Sbom

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Merge all properties of two or more SBOM files",
    invoke_without_command=True,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
    subcommand_metavar="",
)

console = Console()


class HopprMergeJobsPanel(HopprJobsPanel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress(
        "{task.description}",
        SpinnerColumn(),
        "{task.fields[status]}",
    )

    def add_task(self, description: str, **fields) -> int:
        return super().add_task(description, total=1, **fields)


class HopprMergeLayout(HopprLayout):
    """
    Layout of the `hopctl merge` console application
    """

    name: str = "root"
    jobs_panel = HopprMergeJobsPanel()


def manifest_callback(manifest_file: Path | None) -> Path | None:
    """
    Extract and return SBOM refs from manifest files
    """
    if manifest_file is None:
        return None

    # Load manifest file to populate `Sbom.loaded_sboms`
    Manifest.load(manifest_file)

    return manifest_file


def output_file_callback(output_file: Path | None) -> Path:
    """
    Auto-generate an output file name if not provided
    """
    return output_file or Path.cwd() / f"hopctl-merge-{time.strftime('%Y%m%d-%H%M%S')}.json"


def sbom_callback(sbom_sources: list[str | Path]) -> list[str | Path]:
    """
    Load SBOM input files
    """
    # Combine SBOM files from all CLI input file arguments
    _load_sbom_files(sbom_sources)

    return sbom_sources


def sbom_dir_callback(sbom_dirs: list[Path]) -> list[Path]:
    """
    Load SBOM files from input directories
    """
    for sbom_dir in sbom_dirs:
        sbom_files = sbom_dir.glob("*.json")

        _load_sbom_files(list(sbom_files))

    return sbom_dirs


def _load_sbom_files(sbom_files: list[str | Path]) -> None:
    """
    Load SBOM input files
    """
    for sbom_file in sbom_files:
        try:
            Sbom.load(sbom_file)
        except ValidationError as ex:
            raise BadParameter(f"'{sbom_file}' is not a valid SBOM file") from ex


def _merge_sboms() -> Sbom:
    """
    Merge SBOMs into single object
    """
    merged_sbom = Sbom()  # pyright: ignore
    layout = HopprMergeLayout()
    merged_sbom.subscribe(observer=layout, callback=layout.print)

    summary_table = Table("File Name", "# Components", "Elapsed Time", box=None, pad_edge=False, show_edge=False)

    total_components = 0
    for sbom_ref, sbom in Sbom.loaded_sboms.items():
        total_components += len(sbom.components)
        layout.add_job(description=Path(str(sbom_ref)).name, status="[yellow]:pause_button:", sbom=sbom)

    progress_task = layout.overall_progress.progress_bar.tasks[0]
    progress_task.total = sum((task.total or 0) for task in layout.jobs_panel.progress_bar.tasks)

    with Live(layout, console=console, refresh_per_second=10):
        layout.overall_progress.progress_bar.start_task(progress_task.id)

        for task in layout.jobs_panel.progress_bar.tasks:
            layout.print(f"Merging {task.description}...")
            layout.jobs_panel.progress_bar.start_task(task.id)

            merged_sbom.merge(task.fields["sbom"])

            layout.jobs_panel.progress_bar.stop_task(task.id)

            layout.jobs_panel.progress_bar.update(task.id, status="[green]:heavy_check_mark:")
            layout.jobs_panel.progress_bar.advance(task.id)
            layout.overall_progress.progress_bar.advance(progress_task.id)

            summary_table.add_row(task.description, str(len(task.fields["sbom"].components)), f"{task.elapsed:.3f}")

        layout.print("Writing merged SBOM...")

        layout.overall_progress.progress_bar.stop_task(progress_task.id)

    before_after = Table.grid(padding=(0, 1))
    before_after.add_row()
    before_after.add_row("Total components before merge:", str(total_components))
    before_after.add_row("Total components after merge:", str(len(merged_sbom.components)))

    console.print(
        Panel(
            renderable=Group(summary_table, before_after),
            title="[bold blue]Summary",
            title_align="left",
            border_style="purple",
        )
    )

    merged_sbom.unsubscribe(observer=layout)
    return merged_sbom


@app.callback()
def merge(  # pylint: disable=too-many-arguments, unused-argument
    manifest_file: Path = Option(
        None,
        "-m",
        "--manifest",
        callback=manifest_callback,
        dir_okay=False,
        exists=True,
        help="Manifest file containing SBOMs to merge",
        resolve_path=True,
        show_default=False,
    ),
    sbom_files: list[Path] = Option(
        [],
        "-s",
        "--sbom",
        callback=sbom_callback,
        dir_okay=False,
        exists=True,
        help="SBOM file to merge (can be specified multiple times)",
        resolve_path=True,
        show_default=False,
    ),
    sbom_dirs: list[Path] = Option(
        [],
        "-d",
        "--sbom-dir",
        callback=sbom_dir_callback,
        exists=True,
        file_okay=False,
        help="Directory containing SBOM files to merge (can be specified multiple times)",
        resolve_path=True,
        show_default=False,
    ),
    sbom_urls: list[str] = Option(
        [],
        "-u",
        "--sbom-url",
        callback=sbom_callback,
        help="URL of SBOM to merge (can be specified multiple times)",
        metavar="URL",
        show_default=False,
    ),
    output_file: Path = Option(
        None,
        "-o",
        "--output-file",
        callback=output_file_callback,
        dir_okay=False,
        exists=False,
        help=f"Path to output file {typer.style(text='[default: hopctl-merge-YYMMDD-HHMMSS.json]', dim=True)}",
        resolve_path=True,
        show_default=False,
    ),
    deep_merge: bool = Option(
        False,
        "--deep-merge",
        help="Resolve and expand `externalReferences` in-place",
        show_default=False,
    ),
    flatten: bool = Option(
        False,
        "--flatten",
        help="Flatten nested `components` into single unified list",
        show_default=False,
    ),
):
    """
    Merge SBOM files
    """
    if not any([manifest_file, *sbom_dirs, *sbom_files, *sbom_urls]) or len(Sbom.loaded_sboms) < 2:
        raise BadParameter(
            "A minimum of two SBOM files must be provided via the "
            "--sbom, --sbom-dir, --sbom-url, and --manifest arguments."
        )

    CycloneDXBaseModel.deep_merge = deep_merge
    CycloneDXBaseModel.flatten = flatten

    merged_sbom = _merge_sboms()

    output_file.write_text(
        data=merged_sbom.json(exclude_none=True, exclude_unset=True, by_alias=True, indent=2),
        encoding="utf-8",
    )



"""
Collector plugin for docker images
"""

# pylint: disable=too-many-locals

from __future__ import annotations

import re
import urllib.parse
from packageurl import PackageURL

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectDockerPlugin(SerialCollectorPlugin):
    """
    Collector plugin for docker images
    """

    supported_purl_types = ["docker"]
    required_commands = ["skopeo"]
    products: list[str] = ["docker/*"]
    system_repositories: list[str] = ["https://docker.io/"]
    process_timeout = 300

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        self.required_commands = (self.config or {}).get("skopeo_command", self.required_commands)

    def get_image(self, url: str, purl: PackageURL, image_name: str) -> RepositoryUrl:
        """
        Return the image details for skopeo to process

        Args:
            url (str): Repository URL
            purl (PackageURL): Purl of component to operate on
            image_name (str): Name of image with tag

        Returns:
            RepositoryUrl: Image information
        """
        source_image = RepositoryUrl(url=url) / (purl.namespace or "") / urllib.parse.quote_plus(image_name)
        if source_image.scheme != "docker":
            source_image = RepositoryUrl(url="docker://" + re.sub(r"^(.*://)", "", str(source_image)))
        return source_image

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = hoppr.utils.get_package_url(comp.purl)
        version_path = purl.version

        # Determine if purl version contains SHA string, determine proper formatting for skopeo command
        if re.search(r"^(sha256:)?[a-fA-F0-9]{12,64}$", purl.version) is None:
            image_name = f"{purl.name}:{purl.version}"

        elif purl.version.startswith("sha256:"):
            image_name = f"{purl.name}@{purl.version}"
        else:
            image_name = f"{purl.name}@sha256:{purl.version}"
            version_path = f"sha256:{purl.version}"

        version_path = urllib.parse.quote_plus(re.sub(r'^https?://', '', version_path))
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        target_path = target_dir / f"{purl.name}@{version_path}"

        source_image = self.get_image(url=repo_url, purl=purl, image_name=image_name)

        self.get_logger().info(msg="Copying docker image:", indent_level=2)
        self.get_logger().info(msg=f"source: {source_image}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_path}", indent_level=3)

        command = [self.required_commands[0], "copy"]

        password_list = []

        if creds is not None:
            password_list = [creds.password.get_secret_value()]
            command.extend(["--src-creds", f"{creds.username}:{creds.password.get_secret_value()}"])

        if re.match("^http://", repo_url):
            command = [*command, "--src-tls-verify=false"]

        command = [*command, urllib.parse.unquote(str(source_image)), f"docker-archive:{target_path}"]

        proc = self.run_command(command, password_list)

        if proc.returncode != 0:
            msg = f"Skopeo failed to copy docker image to {target_path}, " + f"return_code={proc.returncode}"
            self.get_logger().debug(msg=msg, indent_level=2)

            if target_path.exists():
                self.get_logger().info(msg="Artifact collection failed, deleting file and retrying", indent_level=2)
                target_path.unlink()

            return Result.retry(message=msg)

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""
CycloneDX data models
"""
from __future__ import annotations

import functools
import sys
import uuid

from pathlib import Path
from typing import Annotated, Any, Callable, ClassVar, Iterator, Literal, MutableMapping, TypeAlias, TypeVar

import hoppr_cyclonedx_models.cyclonedx_1_4

from pydantic import BaseModel, Extra, Field, create_model, root_validator, validator
from rapidfuzz import fuzz
from requests import HTTPError

import hoppr.net
import hoppr.utils

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.base import HopprBaseModel
from hoppr.models.types import LocalFile, OciFile, UrlFile

DictStrAny: TypeAlias = dict[str, Any]

__all__ = []

AnyCycloneDXModel = TypeVar("AnyCycloneDXModel", bound="CycloneDXBaseModel")
SbomRef = Annotated[LocalFile | OciFile | UrlFile, Field(..., description="Reference to a local or remote SBOM file")]
SbomRefMap = Annotated[MutableMapping[SbomRef, "Sbom"], Field(default=...)]
UniqueIDMap = Annotated[MutableMapping[str, AnyCycloneDXModel], Field(default={})]

FUZZY_MATCH_THRESHOLD = 85


class CycloneDXBaseModel(BaseModel):
    """
    Base CycloneDX data model
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config options for CycloneDXBaseModel"

    # Defining as ClassVar to allow dynamic model creation using custom root types
    deep_merge: ClassVar[bool] = False
    flatten: ClassVar[bool] = False
    observers: ClassVar[dict[object, Callable]] = {}
    unique_id_map: ClassVar[UniqueIDMap]

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return hash(self.unique_id_callback())

    def __init__(self, **data):
        super().__init__(**data)

        unique_id = self.unique_id_callback()
        type(self).unique_id_map[unique_id] = self

    def _has_field(self, field_name: str) -> bool:
        return hasattr(self, field_name) and (getattr(self, field_name) or field_name in list(self.__fields_set__))

    @staticmethod
    def _is_list_of(obj_type: type, field: Any) -> bool:
        """
        Checks if the provided field is a list of the specified type
        """
        return isinstance(field, list) and all(isinstance(item, obj_type) for item in field)

    def _merge_field(self, target_field_name: str, source_field: Any) -> None:
        """
        Merges `source_field` into the field referenced by `target_field_name`
        """
        merged_field = getattr(self, target_field_name)
        field_type = type(merged_field)

        if isinstance(field_type, type(BaseModel)):
            merged_field = CycloneDXBaseModel.create(model=merged_field)
            source_field = CycloneDXBaseModel.create(model=source_field)
            merged_field.merge(source_field)
        elif self._is_list_of(BaseModel, merged_field):
            merged_field = hoppr.utils.dedup_list([CycloneDXBaseModel.create(model=_field) for _field in merged_field])
            source_field = [CycloneDXBaseModel.create(model=_field) for _field in source_field or []]
            self._merge_field_items(merged_field, source_field)

        setattr(self, target_field_name, merged_field)

    @staticmethod
    def _merge_field_items(target_field: list[CycloneDXBaseModel], source_field: list[CycloneDXBaseModel]) -> None:
        """
        Merges the items from `source_field` into `target_field`
        """
        for source_item in source_field:
            if source_item in target_field:
                merged_item = target_field[target_field.index(source_item)]
                merged_item.merge(source_item)
            else:
                target_field.append(source_item)

    def unique_id_callback(self) -> str:
        """
        Default callback method to get a model object's unique ID
        """
        try:
            callback = {
                "Advisory": lambda obj: obj.url,
                "Commit": lambda obj: obj.uid,
                "Component": lambda obj: obj.bom_ref,
                "Copyright": lambda obj: obj.text,
                "Dependency": lambda obj: obj.ref.__root__,
                "ExternalReference": lambda obj: f"{obj.type}-{obj.url}",
                "License": lambda obj: obj.id or obj.name or repr(self),
                "LicenseChoice": lambda obj: obj.license or obj.expression or repr(self),
                "Note": lambda obj: obj.text,
                "Reference": lambda obj: obj.id,
                "RefType": lambda obj: obj.__root__,
                "Sbom": lambda obj: obj.serialNumber,
                "Service": lambda obj: obj.bom_ref or f"{obj.name}-{obj.version}",
                "Signer": lambda obj: obj.value,
                "Vulnerability": lambda obj: obj.id or repr(self),
            }[type(self).__name__]

            return callback(self)
        except KeyError:
            return repr(self)

    @classmethod
    def create(cls: type[AnyCycloneDXModel], model: BaseModel) -> AnyCycloneDXModel:
        """
        Update a BaseModel object with CycloneDXBaseModel attributes and methods

        Args:
            model (BaseModel): The BaseModel object to update

        Returns:
            AnyCycloneDXModel: The updated BaseModel object
        """
        model_cls = cls.make_model(name=type(model).__name__)
        return model_cls(**model.dict(by_alias=True, exclude_none=True, exclude_unset=True))  # pyright: ignore

    @classmethod
    @functools.cache
    def make_model(cls: type[AnyCycloneDXModel], name: str) -> type[AnyCycloneDXModel]:
        """
        Dynamically create a model class suitable for merging

        Args:
            name (str): Name of the existing model

        Returns:
            type[AnyCycloneDXModel]: The generated model class
        """
        # Return explicitly defined models directly
        if name in {"Component", "ExternalReference", "Hash", "License", "LicenseChoice", "Property"}:
            return sys.modules[__name__].__dict__[name]

        model_cls = hoppr_cyclonedx_models.cyclonedx_1_4.__dict__[name]

        merge_model = create_model(model_cls.__name__, __base__=(cls, model_cls), __module__=__name__)

        # Set model's pydantic `Config` class and `__hash__` method
        setattr(merge_model, "__config__", cls.Config)

        # Add `unique_id_map` class attribute for caching model objects
        merge_model.__class_vars__.add("unique_id_map")
        merge_model.__annotations__["unique_id_map"] = "ClassVar[UniqueIDMap]"
        setattr(merge_model, "unique_id_map", {})

        # Add updated model to current module and make importable from other modules
        setattr(sys.modules[__name__], merge_model.__name__, merge_model)
        __all__.append(merge_model.__name__)
        merge_model.update_forward_refs()

        return merge_model

    @classmethod
    def find(cls: type[AnyCycloneDXModel], unique_id: str) -> AnyCycloneDXModel | None:
        """
        Look up model object by its unique ID string

        Args:
            unique_id (str): Unique ID string to look up

        Returns:
            AnyCycloneDXModel | None: Model object if found, otherwise None
        """
        return cls.unique_id_map.get(unique_id)

    def merge(self: AnyCycloneDXModel, other: AnyCycloneDXModel) -> None:
        """
        Merge model instance of same type into self

        Args:
            other (AnyCycloneDXModel): Model object to merge
        """
        if (self_type := type(self).__name__) != (other_type := type(other).__name__):
            raise TypeError(f"Type '{other_type}' cannot be merged into '{self_type}'")

        self.notify(data=f"  Merging '{type(self).__qualname__}' attributes...")

        for field_name in self.__fields__:
            self.notify(data=f"    Merging field '{type(self).__qualname__}.{field_name}'...")

            source_field = getattr(other, field_name, None)

            if source_field is None:
                continue

            if not self._has_field(field_name):
                setattr(self, field_name, source_field)
            else:
                self._merge_field(field_name, source_field)

    def notify(self, data: str) -> None:
        """
        Call the callback function for all registered subscribers
        """
        for callback in self.observers.values():
            callback(data)

    def subscribe(self, observer: object, callback: Callable) -> None:
        """
        Register an observer
        """
        self.observers[observer] = callback

    def unsubscribe(self, observer: object) -> None:
        """
        Unregister an observer
        """
        self.observers.pop(observer, None)


def _extract_components(components: list[Component]) -> list[Component]:
    """
    Explicitly set scope of flattened components to `exclude`
    """
    for component in components:
        setattr(component, "scope", "excluded")

    return hoppr.utils.dedup_list(components)


def _extract_sbom_components(external_refs: list[ExternalReference]) -> list[Component]:
    """
    Extracts `external_refs` of type "bom" and returns the set of their components
    """
    components: list[Component] = []

    for ref in _get_bom_refs(external_refs):
        sbom = Sbom.load(_resolve_sbom_source(ref.url))
        sbom.components = _extract_components(sbom.components)
        components.extend(sbom.components)
        external_refs.remove(ref)

    return components


def _flatten_component(component: Component) -> list[Component]:
    """
    Helper function to flatten a component's subcomponents into a set
    """
    flattened = []

    for subcomponent in component.components or []:
        # Ensure validator is run to set `bom_ref`
        subcomponent = Component(**subcomponent.dict())
        setattr(subcomponent, "scope", "excluded")

        # Flatten nested components into top level components list
        flattened.append(subcomponent)

    component.components.clear()
    return flattened


def _get_bom_refs(external_refs: list[ExternalReference]) -> Iterator[ExternalReference]:
    """
    Get `externalReferences` of type "bom"
    """
    yield from (ref.copy(deep=True) for ref in (external_refs or []) if ref.type == "bom")


def _resolve_sbom_source(source: str) -> str | Path | DictStrAny:
    """
    Resolves an SBOM source as a file path, URL or `dict`
    """
    return Path(source.removeprefix("file://")).resolve() if source.startswith("file://") else source


def _validate_components(cls, components: list[Component]) -> list[Component]:
    """
    Validator to optionally flatten `components` list
    """
    if not cls.flatten:
        return components

    flattened = list(components)

    for component in components:
        flattened.extend(_flatten_component(component))

    return hoppr.utils.dedup_list(flattened)


def _validate_external_refs(cls, external_refs: list[ExternalReference], values: DictStrAny) -> list[ExternalReference]:
    """
    Validator to optionally resolve `externalReferences`
    """
    external_refs = [ExternalReference.create(ref) for ref in external_refs or []]

    if cls.deep_merge:
        external_ref_components = _extract_sbom_components(external_refs)
        values["components"] = hoppr.utils.dedup_list([*values.get("components", []), *external_ref_components])

    return external_refs


class ExternalReference(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.ExternalReference):
    """
    ExternalReference data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for ExternalReference model"
        extra = Extra.allow

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Hash(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.Hash):
    """
    Hash data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Hash model"
        extra = Extra.allow

    content: str = Field(  # type: ignore[assignment]
        default=...,
        examples=["3942447fac867ae5cdb3229b658f4d48"],
        title="Hash Content (value)",
        regex=r"^([a-fA-F0-9]{32}|[a-fA-F0-9]{40}|[a-fA-F0-9]{64}|[a-fA-F0-9]{96}|[a-fA-F0-9]{128})$",
    )

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class License(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.License):
    """
    License data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for License model"
        extra = Extra.allow

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseChoice(CycloneDXBaseModel):
    """
    LicenseChoice data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for LicenseChoice model"
        extra = Extra.allow

    license: License | None = None
    expression: str | None = Field(
        default=None,
        examples=["Apache-2.0 AND (MIT OR GPL-2.0-only)", "GPL-3.0-only WITH Classpath-exception-2.0"],
        title="SPDX License Expression",
    )

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Property(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.Property):
    """
    Property data model derived from CycloneDXBaseModel
    """

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}


class Component(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.Component):
    """
    Component data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Component model"
        extra = Extra.allow

    bom_ref: str = Field(default=None, alias="bom-ref")  # type: ignore[assignment]
    components: list[Component] = Field(default=[])  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = Field(default=[])  # type: ignore[assignment]
    hashes: list[Hash] = Field(default=[])  # type: ignore[assignment]
    licenses: list[LicenseChoice] = Field(default=[])  # type: ignore[assignment]
    properties: list[Property] = Field(default=[])  # type: ignore[assignment]

    # Attributes not included in schema
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components: classmethod = validator("components", allow_reuse=True, always=True)(_validate_components)
    validate_external_refs: classmethod = validator("externalReferences", allow_reuse=True)(_validate_external_refs)

    def __eq__(self, other: object) -> bool:
        try:
            if not isinstance(other, Component):
                raise TypeError

            self_purl = hoppr.utils.get_package_url(self.purl)
            other_purl = hoppr.utils.get_package_url(other.purl)
        except (TypeError, ValueError):
            return super().__eq__(other)

        qual_keys = hoppr.utils.dedup_list({**self_purl.qualifiers, **other_purl.qualifiers})

        def _qualifier_match(key: str) -> bool:
            # Compare only if both purls have a value for specified qualifier
            if key in self_purl.qualifiers and key in other_purl.qualifiers:
                return (
                    fuzz.ratio(self_purl.qualifiers.get(key, ""), other_purl.qualifiers.get(key, ""))
                    > FUZZY_MATCH_THRESHOLD
                )

            return True

        return all(
            [
                self_purl.name == other_purl.name,
                self_purl.type == other_purl.type,
                self_purl.namespace == other_purl.namespace,
                str(self_purl.version).removeprefix("v") == str(other_purl.version).removeprefix("v"),
                self_purl.subpath == other_purl.subpath,
                *[_qualifier_match(key) for key in qual_keys],
            ]
        )

    @root_validator(allow_reuse=True, pre=True)
    @classmethod
    def validate_component(cls, values: DictStrAny) -> DictStrAny:
        """
        Validator to set a Component's `bom-ref` identifier if not set
        """
        if not any([values.get("purl"), values.get("bom-ref"), all([values.get("name"), values.get("version")])]):
            raise ValueError(
                "Either 'bom-ref' or 'purl' must be defined, or 'name' and 'version' must be defined on a component"
            )

        def _unescape_unicode(value: str) -> str:
            # Decode any unicode escape sequences, e.g. "\u0026" -> "&"
            return value.encode(encoding="utf-8").decode()

        if values.get("purl"):
            values["purl"] = _unescape_unicode(str(values["purl"]))

        bom_ref = str(
            values.get("purl")
            or f"{'@'.join(filter(None, [values.get('name'), values.get('version')]))}"
            or values.get("bom-ref")
        )

        values["bom-ref"] = _unescape_unicode(bom_ref)

        return values


Component.update_forward_refs()


class Sbom(CycloneDXBaseModel, hoppr_cyclonedx_models.cyclonedx_1_4.CyclonedxSoftwareBillOfMaterialsStandard):
    """
    Sbom data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Sbom model"

    bomFormat: Literal["CycloneDX"] = "CycloneDX"  # type: ignore[assignment]
    components: list[Component] = Field(default=[])  # type: ignore[assignment]
    externalReferences: list[ExternalReference] = Field(default=[])  # type: ignore[assignment]
    serialNumber: str = Field(
        default="",
        regex=r"^urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
    )
    specVersion: str = "1.4"
    version: int = 1

    # Attributes not included in schema
    field_schema: Literal["http://cyclonedx.org/schema/bom-1.4.schema.json"] | None = Field(
        default=None, exclude=True, alias="$schema"
    )
    loaded_sboms: ClassVar[SbomRefMap] = {}
    unique_id_map: ClassVar[UniqueIDMap] = {}

    validate_components: classmethod = validator("components", allow_reuse=True, always=True)(_validate_components)
    validate_external_refs: classmethod = validator("externalReferences", allow_reuse=True)(_validate_external_refs)

    @validator("serialNumber", allow_reuse=True, always=True, pre=True)
    @classmethod
    def validate_serial_number(cls, serial_number: str | None) -> str:
        """
        Validator to generate SBOM serial number if None
        """
        return serial_number or uuid.uuid4().urn

    @classmethod
    def find_ref(cls, ref_type: Literal["local", "oci", "url"], location: str | Path) -> Sbom | None:
        """
        Look up SBOM object by reference

        Args:
            ref_type (Literal["local", "oci", "url"]): Type of SBOM reference
            location (str | Path): Location of SBOM reference

        Returns:
            Sbom | None: SBOM object if found, otherwise None
        """
        # pylint: disable=duplicate-code
        match ref_type:
            case "local":
                return cls.loaded_sboms.get(LocalFile(local=Path(location)), None)
            case "oci":
                return cls.loaded_sboms.get(OciFile(oci=str(location)), None)
            case "url":
                return cls.loaded_sboms.get(UrlFile(url=str(location)), None)
            case _:
                return None

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Sbom:
        """
        Load SBOM from local file, URL, or dict
        """
        # pylint: disable=duplicate-code
        match source:
            case dict():
                sbom = cls(**source)
            case Path():
                # Convert source to relative path if in current working directory subpath
                source = source.resolve()
                source = source.relative_to(Path.cwd()) if source.is_relative_to(Path.cwd()) else source

                sbom = cls.parse_file(source)
                cls.loaded_sboms[LocalFile(local=source)] = sbom
            case str():
                try:
                    sbom_dict = hoppr.net.load_url(source)
                    if not isinstance(sbom_dict, dict):
                        raise TypeError("URL SBOM was not loaded as dictionary")

                    sbom = cls.parse_obj(sbom_dict)
                    url_ref = UrlFile(url=source)
                    cls.loaded_sboms[url_ref] = sbom
                except (HopprLoadDataError, HTTPError) as ex:
                    raise HopprLoadDataError from ex

        return sbom
